/* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
/* jshint esnext: true */
/* jshint multistr: true */
/**
    Copyright (C) 2017 Jason Gray <jasonlevigray3@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

const Gio = imports.gi.Gio;
const Gtk = imports.gi.Gtk;
const Lang = imports.lang;
const GnomeSession = imports.misc.gnomeSession;	

const DBusProxy = Gio.DBusProxy.makeProxyWrapper('<node> \
<interface name="org.freedesktop.DBus"> \
  <method name="ListNames">\
    <arg type="as" direction="out" />\
  </method>\
  <signal name="NameOwnerChanged"> \
    <arg type="s" direction="out" name="name" /> \
    <arg type="s" direction="out" name="oldOwner" /> \
    <arg type="s" direction="out" name="newOwner" /> \
  </signal> \
</interface> \
</node>');

const MprisPlayerProxy = Gio.DBusProxy.makeProxyWrapper('<node> \
<interface name="org.mpris.MediaPlayer2.Player"> \
  <method name="Pause" /> \
  <method name="Stop" /> \
  <method name="Play" /> \
  <property name="CanPause" type="b" access="read" />\
  <property name="CanPlay" type="b" access="read" />\
  <property name="PlaybackStatus" type="s" access="read" />\
</interface> \
</node>');

const MPRIS_PREFIX = 'org.mpris';

const Morphine = new Lang.Class({
    Name: 'Morphine',

    _init: function() {
      this._players = [];
      this._dbus = null;
      this._dbusNameOwnerChangedId = null;
      this._presence = null;
      this._statusChangedId = null;
      new DBusProxy(Gio.DBus.session,
                    'org.freedesktop.DBus',
                    '/org/freedesktop/DBus',
                    Lang.bind(this, this._onDBusProxyReady));
    },

    _pausePlayers: function() {
      // I would prefer to just create the proxy once and use it
      // for the life of the player, but players have a nasty
      // habit of not updating properties when they change.
      // This way we get up to date properties every time.
      this._players.forEach(Lang.bind(this, function(player) {
        new MprisPlayerProxy(Gio.DBus.session,
                             player._name,
                             '/org/mpris/MediaPlayer2',
                             Lang.bind(this, function(proxy) {
                               player._paused = false;
                               if (proxy.PlaybackStatus == 'Playing') {       
                                 if (proxy.CanPause) {
                                   player._paused = true;
                                   proxy.PauseRemote();
                                 }
                                 else {
                                   proxy.StopRemote();
                                 }
                               }
                             }));
      }));
    },

    _resumePlayers: function() {
      // Resume playback if the session idle fade out is interrupted and we paused the player.
      // Don't resume if the player was stopped. Stopped is an inconsistent state.
      // It's unclear what will happen if we hit play. Will the playlist start over?
      // Will the song start over? Will it do anything at all? 
      this._players.forEach(Lang.bind(this, function(player) {
        if (player._paused) {
          player._paused = false;
          new MprisPlayerProxy(Gio.DBus.session,
                               player._name,
                               '/org/mpris/MediaPlayer2',
                               Lang.bind(this, function(proxy) {
                                 if (proxy.CanPlay) {       
                                   proxy.PlayRemote();
                                 }
                               }));
        }
      }));
    },

    _onNameOwnerChanged: function(proxy, sender, [name, oldOwner, newOwner]) {
      if (!name.startsWith(MPRIS_PREFIX)) {
        return;
      }
      if (newOwner && !oldOwner) {
        for (let i in this._players) {
          if (name == this._players[i]._name) {
            return;
          }
        }
        this._players.push({_name: name});
      }
      else if (oldOwner && !newOwner) {
        for (let i in this._players) {
          if (name == this._players[i]._name) {
            this._players.splice(i, 1);
            break;
          }
        }
      }
    },

    _onDBusProxyReady: function(proxy) {
      this._dbus = proxy;
      this._dbus.ListNamesRemote(Lang.bind(this, function(names) {
        names[0].forEach(Lang.bind(this, function(name) {
          if (name.startsWith(MPRIS_PREFIX)) {
            this._players.push({_name: name});
          }
        }));
      }));

      this._dbusNameOwnerChangedId = this._dbus.connectSignal('NameOwnerChanged',
                                       Lang.bind(this, this._onNameOwnerChanged));

      this._presence = new GnomeSession.Presence();
      this._statusChangedId = this._presence.connectSignal('StatusChanged',
                                Lang.bind(this, function(proxy, senderName, [status]) {
                                  if (status == GnomeSession.PresenceStatus.IDLE) {
                                    this._pausePlayers();
                                  }
                                  else if (status == GnomeSession.PresenceStatus.AVAILABLE) {
                                    this._resumePlayers();
                                  } 
                                })); 
    },

    destroy: function() {
      if (this._dbusNameOwnerChangedId) {
        this._dbus.disconnectSignal(this._dbusNameOwnerChangedId);
      }
      if (this._statusChangedId) {
        this._presence.disconnectSignal(this._statusChangedId);
      }
      this._players = null;
      this._dbus = null;
      this._dbusNameOwnerChangedId = null;
      this._presence = null;
      this._statusChangedId = null;
    }
});

let _extension = null;

//We don't need to init anything...
function init() {}

function enable() {
    _extension = new Morphine();
}

function disable() {
    if (_extension) {
      _extension.destroy();
      _extension = null;
    }
}
